import os
import logging
from typing import Text, Optional, Tuple

import sqlalchemy.orm

import rasax.community.config
import rasax.community.constants
import rasax.community.utils.common
import rasax.community.utils.cli
from rasax.community.services.user_service import (
    UserService,
    UserException,
    RoleException,
    AuthMechanisms,
    ADMIN,
)
from rasax.community.services.license_service import LicenseService

logger = logging.getLogger(__name__)


def create_user(
    service: UserService,
    username: Text,
    password: Optional[Text],
    role: Text,
    team: Text,
) -> None:
    """Create a user with specified parameters.

    Arguments:
        service: An instance of the `rasax.community.services.user_service.UserService`.
        username: A username for the user.
        password: A password for the user.
        role: A role for the user (must be either "admin", "annotator", "tester", or "guest").
        team: A valid team name for the user.
    """
    try:
        service.create_user(
            username, password, team, role, AuthMechanisms.username_password,
        )
    except UserException as e:
        rasax.community.utils.cli.print_error_and_exit(
            f"User '{e}' already exists. You can update the password by "
            f"running `sudo python rasa_x_commands.py create --update "
            f"admin {e} <new_pw>`"
        )
    except RoleException:
        rasax.community.utils.cli.print_error_and_exit(
            f"Role '{role}' does not exist. Please select a valid role."
        )


def _read_password() -> Optional[Text]:
    password = os.getenv(rasax.community.constants.ENV_RASA_X_PASSWORD)

    if not password and (
        rasax.community.config.LOCAL_MODE
        or not LicenseService.is_enterprise_activated()
    ):
        # Generate a password instead when either:
        # - Enterprise is not activated.
        # - We are running on Local mode.
        password = rasax.community.utils.common.random_password()

    return password


def setup_initial_admin_user(
    session: sqlalchemy.orm.Session, create: bool
) -> Tuple[Optional[Text], Optional[Text]]:
    """Setup the initial administrator user for this server.

    The initial administrator user will only be created if `create` is
    `True`. How the passoword is set depends on whether we are running on Local
    or Server mode, and if Enterprise is activated or not.

    This function is a merging of the old "community" and Enterprise
    initialisation functions. It is also used for Local mode initialisation.

    Args:
        session: SQLAlchemy session.
        create: Whether we should attempt to create the administrator user.

    Returns:
        See return value of `_set_initial_admin_user_password`.
    """
    user_service = UserService(session)
    username = os.getenv(
        rasax.community.constants.ENV_RASA_X_USERNAME,
        rasax.community.constants.DEFAULT_USERNAME,
    )

    # First, if required, create the user
    if create:
        try:
            user = user_service.create_user(
                username, None, rasax.community.config.team_name, ADMIN
            )
            if user:
                rasax.community.utils.cli.print_info(
                    f"Created initial administrator user '{username}'."
                )
        except UserException:
            logger.debug(
                f"Skip creating initial admin user '{username}', as it already exists."
            )

    # Second, decide if we set or update its password or not
    return _set_initial_admin_user_password(user_service, username)


def _set_initial_admin_user_password(
    user_service: UserService, username: Text
) -> Tuple[Optional[Text], Optional[Text]]:
    """Set the initial administrator user's password if possible.

    Args:
        user_service: `UserService` with a session attached to it.
        username: The administrator user's username.

    Returns:
        Tuple of optional username and password. The username will only be
        returned if the initial administrator user exists (even if it was
        created outside this function). The password will only be returned if
        it was set during the call to this function.
    """
    if not user_service.fetch_user(username):
        return None, None

    password = None

    if rasax.community.config.LOCAL_MODE:
        # Local mode: always overwrite password with a new one.
        password = _read_password()

        user_service.insert_or_update_user(
            username, password, rasax.community.config.team_name
        )
    elif user_service.is_user_password_set_required(username):
        # Server mode (only when admin user has no password).
        password = _read_password()

        if password:
            user_service.insert_or_update_user(
                username, password, rasax.community.config.team_name
            )
        else:
            rasax.community.utils.cli.print_warning(
                f"No password set for user '{username}'. Make sure you have specified "
                f"the '{rasax.community.constants.ENV_RASA_X_PASSWORD}' environment "
                f"variable."
            )

    return username, password
